import numpy as np
import os


def running_mean(x, N):
    '''

    Parameters
    ----------
    x : ndarray
        1d array.
    N : int
        Number of elements used to smooth.

    Returns
    -------
    1d array
        Smoothed array.

    '''
    cumsum = np.nancumsum(np.insert(x, 0, 0))
    return (cumsum[N:] - cumsum[:-N]) / float(N)


def nanravel(matrix):
    '''
    Return a contiguous flattened array without nan.

    Parameters
    ----------
    matrix : ndarray
        Input array.

    Returns
    -------
    matrix : 1d array
        Flattened array.

    '''
    matrix = np.ravel(matrix)
    matrix = matrix[np.logical_not(np.isnan(matrix))]
    return matrix

def listdir(path, suffix, suffix_on=True, prefix_on=False):
    '''
    Return a list containing the names of the files in the directory.
    Filter path with specific suffix(or prefix). This is a improved version of os.listdir

    Parameters
    ----------
    path : string
        Input path.
    suffix : string
        Suffix or prefix.
    suffix_on : bool, optional
        Whether to use suffix. The default is 1.
    prefix : TYPE, optional
        Whether to use suffix. The default is 0.

    Returns
    -------
    ls : list
        A list containing the names of the files in the directory with specific suffix or prefix.

    '''
    tem = os.listdir(path)
    if prefix_on == True:
        num = len(suffix)
        ls = np.array([])
        for p in tem:
            if p[:num] == suffix:
                ls = np.append(ls, p)
    else:
        if suffix_on == True:
            num = len(suffix)
            ls = np.array([])
            for p in tem:
                if p[-num:] == suffix:
                    ls = np.append(ls, p)

    return ls



def nanaverage(data,weights=0,axis=9999):
    '''
    Weighted mean for array with nan.

    Parameters
    ----------
    data : ndarray
        Input data.
    weights : nadarry, optional
        DESCRIPTION. The default is 0.
    axis : TYPE, optional
        DESCRIPTION. The default is 100.

    Raises
    ------
    ValueError
        the data and weights are expected to have the same shape and nan values.

    Returns
    -------
    TYPE
        Weighted mean.

    '''
    axis = np.array([axis])
    if np.all(axis == 9999):
        axis = np.arange(len(data.shape))
    if np.all(weights==0):
        weights = np.copy(data)
        weights[:]=1
    if data.shape != weights.shape:
        raise ValueError('data and weights have different shape')
    if len(nanravel(weights)) != len(nanravel(data)):       
        weights[np.isnan(data)]=np.nan
        print('Length different! Assign nan to weights where data eq nan')
    if len(nanravel(weights)) != len(nanravel(data)):
        raise ValueError('data and weights have different nan')
    tem1 = np.nansum(data*weights,axis=tuple(axis))
    tem2 = np.nansum(weights,axis=tuple(axis))
    return tem1/tem2


def arrayinfo(array,detail=False):
    '''
    Print array info for ndarray.

    Parameters
    ----------
    array : ndarray
        Input array.
    detail : bool, optional
        Whether to print detailed info of the array. The default is False.

    Returns
    -------
    None.

    '''
    if type(array) != np.ndarray:
        print("It's not an numpy.ndarray")
        return None
    array_max = np.max(array)
    if np.isnan(array_max):
        print('Nan exists\nmaxium:{:.2f}'.format(np.nanmax(array)))
        print('minmium:{:.2f}'.format(np.nanmin(array)))
    else:
        print('Nan not exists\nmaxium:{:.2f}'.format(array_max))
        print('minmium:{:.2f}'.format(np.min(array)))
    print('shape:', array.shape)
    print('dtype:', array.dtype)
    if detail:
        print('Unique value:',np.unique(array))
        print('Mean:',np.nanmean(array))
        print('Std:',np.nanstd(array))




def exclude_outlier(in_array):
    '''
    Exclude outlier (3 sigma)

    Parameters
    ----------
    array : 1d array
        Input data.

    Returns
    -------
    array : 1d array
        Output data with outlier=np.nan.

    '''
    array = np.copy(in_array)
    arraystd = np.nanstd(array, ddof=1)
    arraymean = np.nanmean(array)
    arrayoutlier = np.where(np.abs(array - arraymean) > (3 * arraystd))
    print(array[arrayoutlier])
    print(arraystd)
    array[arrayoutlier] = np.nan
    return array

'''
220422 
It seems that the outlier will be eventually excluded in a few loops
n = np.zeros((1000))
for i in range(1000):
    aa = np.random.normal(100,10,20000)
    bb = exclude_outlier(aa)
    while np.nanstd(aa, ddof=1) != np.nanstd(bb, ddof=1):
        aa = bb
        bb = exclude_outlier(aa)
        n[i] = n[i]+1
'''

def r2a(rasterfn,dtype = None):
    from osgeo import gdal
    '''
    Convert raster to array

    Parameters
    ----------
    rasterfn : String
        raster file path.

    Returns
    -------
    2d-ndarray
        numpy array from raster band 1.

    '''
    if dtype==None:
        raster = gdal.Open(rasterfn)
        band = raster.GetRasterBand(1)
    else:
        raster = gdal.Open(rasterfn)
        band = raster.GetRasterBand(1).astype(dtype)
    return band.ReadAsArray()

def n2a(ncfile,var_name):
    '''
    Convert nc to array

    Parameters
    ----------
    ncfile : String
        NC file path.
    var_name : String
        Variable name.

    Returns
    -------
    array : TYPE
        DESCRIPTION.

    '''
    from netCDF4 import Dataset
    nc = Dataset(ncfile, 'r')
    array = nc.variables['{}'.format(var_name)][:]
    return array



# currently not used 
def func_replace(filein_func, name_func, idx_func, value_func):
  import re
  with open(filein_func, 'r') as fin_func:
    lines_func = fin_func.readlines()
    flag = 0
    for i_func, line_func in enumerate(lines_func):
      if re.match(name_func, line_func):
        lines_func[i_func] = line_func[:idx_func] + str(value_func) +'\n'
        flag += 1
    if flag==0:
      print('Cannot find %s'%name_func)
  fin_func.close()
  with open(filein_func, 'w') as fin_func:
    fin_func.writelines(lines_func)
  fin_func.close()


def func_to_landpoint(lat, lon, res=0.5):
  '''
  This is to convert real lat, lon to landpoint value in CRU-NCEP forcing file.
  ilat, ilon mean python index starting from 0 [90, -180].
  landpoint starting from 1.
  '''
  ilat = int((90-lat) / res)
  ilon = int((lon+180) / res)
  landpoint = int(ilat*360/res + ilon + 1)
  return landpoint

def func_to_ilatilon(landpoint, res=0.5):
    '''
    This is to convert landpoint value in CRU-NCEP forcing file to index of lat and lon
    ilat, ilon mean python index starting from 0 [90, -180].
    landpoint starting from 1.
    '''
    ilat = int((landpoint-1) / (360/res))
    ilon = int(landpoint-ilat*360/res-1)
    return ilat, ilon

def func_to_latlon(landpoint, res=0.5):
    '''
    This is to convert landpoint value in CRU-NCEP forcing file to lat and lon
    ilat, ilon mean python index starting from 0 [90, -180].
    landpoint starting from 1.
    '''
    ilat = int((landpoint-1) / (360/res))
    lat = 90 - res*ilat
    ilon = int(landpoint-ilat*360/res-1)
    lon = 180 + res*ilon
    return lat, lon



























import numpy as np
import matplotlib.pyplot as plt
def globalarea(reso,lat = [90, -90],lon=[-180, 180]):
    '''
    Return global or regional area for each grid cell

    Parameters
    ----------
    reso : float
        Resolution of the area (should not be too small or there will be momery error).
    lat : list, optional
        Latitude of interest. The default is [90, -90] (global).
    lon : list, optional
        Longitude of interest. The default is [-180, 180] (global).
        Some frequently-used area: 
            China: lat = [55,15],lon=[70,140]
            SA: lat = [15,-60],lon=[-90,-150]
            Africa: lat = [30,-40],lon=[-150,55]

    Returns
    -------
    2d-array
        Area for each grid cell.

    '''
    lat = np.array(lat)
    lon = np.array(lon)
    if np.any(lat > 90) or np.any(lat < -90):
        raise Exception('Error: latitude out of bound.')
    if np.any(lon > 180) or np.any(lon < -180):
        raise Exception('Error: longitude out of bound.')
    nlon = int(360 / reso)
    nlat = int(180 / reso)
    lat_area = np.zeros(nlat)
    RE = 6371.393  #earth redius 6371.393 k
    for i in np.arange(nlat):
        lat_area[i] = abs(2 * np.pi * (RE**2) *
            (np.cos(np.pi * (i + 1) / nlat) - np.cos(np.pi * i / nlat))) / nlon
    area = np.zeros((nlat, nlon))
    for i in np.arange(nlon):
        area[:, i] = lat_area
    return (area[int((90 - lat[0]) *(1 / reso)):int((90 - lat[1]) * (1 / reso)),
                 int((180 + lon[0]) *(1 / reso)):int((180 + lon[1]) * (1 / reso))])


def gridarea(reso,lat,lon,size=1):
    '''
    Parameters
    ----------
    reso : float
        Resolution of the area.
    lat : float
        Upper latitude.
    lon : float
        Left longitude.
    size : Float, optional
        The number of degree for creating area. Unit: degree.The default is 1.

    Raises
    ------
    Exception
        resolution should be smaller than 1.

    Returns
    -------
    2d-array
        Area for each grid cell.

    '''
    if reso>1:
        raise Exception('Error: resolution > 1.')
    RE = 6371.393  #earth redius 6371.393 k
    nsize = int(size/reso)
    lat_area = np.zeros(nsize)
    for i in np.arange(nsize):
        lat_area[i] = abs(2 * np.pi * (RE**2) *
            (np.sin(np.pi * (lat - i * reso - reso) / 180) -
             np.sin(np.pi * (lat - i * reso) / 180))) * reso / 360
    area = np.zeros((nsize, nsize))
    for i in np.arange(nsize):
        area[:, i] = lat_area
    return (area)






        



import gloce.func
import gloce.area
import gloce.interpolation
import gloce.resample
#----func----
running_mean = gloce.func.running_mean
nanravel = gloce.func.nanravel
listdir = gloce.func.listdir
nanaverage = gloce.func.nanaverage
arrayinfo = gloce.func.arrayinfo
exclude_outlier = gloce.func.exclude_outlier
r2a = gloce.func.r2a
n2a = gloce.func.n2a
func_to_landpoint = gloce.func.func_to_landpoint
func_to_ilatilon = gloce.func.func_to_ilatilon
func_to_latlon = gloce.func.func_to_latlon
#------------

#----resample----
downscaling = gloce.resample.downscaling
upscaling = gloce.resample.upscaling
#----------------

#----area----
globalarea = gloce.area.globalarea
gridarea = gloce.area.gridarea
#------------

#----interpolation----
interpolation_map = gloce.interpolation.interpolation_map
interpolation_time = gloce.interpolation.interpolation_time
#---------------------


import numpy as np
def interpolation_map(map_in,min_obs=10,mask = np.nan):
    '''
    Interpolation with the nearest value on the map
    
    Parameters
    ----------
    map_in : 2d array
        map that need to be interpolated 
    min_obs : int, optional
        the minimum pixels used to interpolate the grid cell
    mask : 2d array, optional
        only execuate interpolation where mask>0.
    Returns
    -------
    smy : 2d array
        Interpolated map.
    '''
    lat = map_in.shape[0]
    lon = map_in.shape[1]
    smy = np.zeros((lat,lon))
    smy[:] = np.nan
    filled_value = np.nansum(map_in*mask)/np.nansum(mask)
    if np.isnan(filled_value):
        raise ValueError('all data is nan, return nan')
    else:
        print('filled_value = {:.2f}'.format(filled_value))
        if np.isnan(mask).all():
            mask=np.ones((lat,lon))
        else:
            print('Mask is used.')
        for i in range(lat):
            print('\r'+'{:.2%}'.format((i+1)/lat),end = '')
            for j in range(lon):
                if mask[i,j]>0:
                    if np.isnan(map_in[i,j]):
                        n = 1
                        data = np.nan
                        while np.isnan(data):
                            xmin = j-n
                            if xmin<0:
                                xmin=0
                            ymin = i-n
                            if ymin<0:
                                ymin=0
                            select = map_in[ymin:i+n+1,xmin:j+n+1]
                            if np.nansum(~np.isnan(select))>=10:
                                data = np.nanmean(select)
                            n=n+1
                            if n>10:
                                data = filled_value
                        smy[i,j] = data
                    else:
                        smy[i,j] = map_in[i,j]
    return smy


def interpolation_time(array, time):
    '''
    Linear interpolation for known time series

    Parameters
    ----------
    array : 2d array
        Shape(nraw,mcol) nraw: N rows of data; mcol:M column data.
    time : list
        A list of time of each colunm; the shape should be the same as M.
        The time step is assumed to be 1 year(maybe change in future)

    Returns
    -------
    smy : 2d array
        Interpolated array.

    '''
    time = np.array(time)
    raws = array.shape[0]
    time = time - time[0]
    smy = np.zeros((raws, (time[-1] + 1)))
    for i in range(len(time) - 1):
        tem = np.zeros((raws, (time[i + 1] - time[i])))
        for j in range(time[i + 1] - time[i]):
            tem[:,
                j] = array[:, i] + (array[:, i + 1] -
                                    array[:, i]) * j / (time[i + 1] - time[i])
        smy[:, time[i]:time[i + 1]] = tem
    smy[:, -1] = array[:, -1]
    return smy









import numpy as np
from scipy import stats


def upscaling_2d(matrix, rstep, cstep, func=np.mean, weights=9999):
    '''
    Resample 2d-array data.
    220602
    If the func is not np.nanmean or stats.mode, 
    the old algorithm from Li Wei is faster actually (4 times faster).
    Parameters
    ----------
    matrix : 2d-array
        Input data.
    rstep : int
        Scaling factor for row.
    cstep : int
        Scaling factor for colunm.
    func : TYPE, optional
        DESCRIPTION. The default is np.mean.
    weights : ndarray, optional
        Weights that should have the same shape as matrix. The default is 9999.
    return_w : bool, optional
        Whethert to return resampled weights. The default is False.

    Returns
    -------
    2d-array
        Resampled data.

    '''
    row, col = np.shape(matrix)
    temp = matrix.reshape(int(row / rstep), rstep, int(col / cstep), cstep)
    if func==np.nanmean or func==stats.mode:
        temp = np.transpose(temp,[1,3,0,2])
        temp = temp.reshape(rstep*cstep,int(row / rstep),int(col / cstep))
        if np.all(weights==9999):
            temp = func(temp, axis=0)
        else:
            weights = weights.reshape(int(row / rstep), rstep, int(col / cstep), cstep)
            weights = np.transpose(weights,[1,3,0,2])
            weights = weights.reshape(rstep*cstep,int(row / rstep),int(col / cstep))
            temp = func(temp, axis=0,weights = weights)
    else: # func = np.sum/np.nansum/np.mean
        if np.all(weights==9999):
            temp = func(temp, axis=1)
            temp = func(temp, axis=2)
        else:
            weights = weights.reshape(int(row / rstep), rstep, int(col / cstep), cstep)
            temp = func(temp, axis=1,weights = weights)
            weights = np.nansum(weights, axis=1)
            temp = func(temp, axis=2,weights = weights)
            #weights = np.nansum(weights, axis=2)
    return temp



def upscaling(matrix, rstep, cstep, func=np.mean, weights=9999,return_w=False):
    '''
    Resample data with dimension less than 5 to reduce resolution.

    Parameters
    ----------
    matrix : ndarray
        Input data.
    rstep : int
        Scaling factor for row.
    cstep : int
        Scaling factor for colunm.
    func : TYPE, optional
        DESCRIPTION. The default is np.mean. (stats.mode for modal number)
    weights : ndarray, optional
        Weights that should have the same shape as matrix. The default is 9999.
    return_w : bool, optional
        Whethert to return resampled weights. The default is False.

    Raises
    ------
    ValueError
        The dimension of input data should be less than 5.

    Returns
    -------
    ndarray
        Resampled data.

    '''
    ndim = np.array(matrix.shape)
    ndim[-1] = ndim[-1]/cstep 
    ndim[-2] = ndim[-2]/rstep 
    if len(ndim)==2:
        temp = upscaling_2d(matrix, rstep, cstep, func=func, weights=weights,return_w=return_w)
    if len(ndim)==3:
        temp = np.zeros(ndim)
        for i in range(ndim[0]):
            temp[i] = upscaling_2d(matrix[i], rstep, cstep, func=func, weights=weights,return_w=return_w)
    if len(ndim)==4:
        temp = np.zeros(ndim)
        for i in range(ndim[0]):
            for j in range(ndim[1]):
                temp[i,j] = upscaling_2d(matrix[i,j], rstep, cstep, func=func, weights=weights,return_w=return_w)
    if len(matrix.shape)>4:
        raise ValueError('The dimension of input data should be less than 5.')
    return temp


def downscaling(matrix, rstep, cstep):
    '''
    220602
    This is much faster than previous version.
    Parameters
    ----------
    matrix : TYPE
        DESCRIPTION.
    rstep : TYPE
        DESCRIPTION.
    cstep : TYPE
        DESCRIPTION.

    Returns
    -------
    temp : TYPE
        DESCRIPTION.

    '''
    row, col = np.shape(matrix)
    temp = np.zeros((row*rstep,col*cstep))
    for i in range(row):
        for j in range(col):
            temp[i*rstep:(i+1)*rstep,j*cstep:(j+1)*cstep]=matrix[i,j]
    return temp

    